from dataclasses import dataclass
from math import cos, degrees, pi, sin
from typing import Callable, Optional, Tuple


@dataclass
class Helix:
    """
    A Helix has radius, pitch and height parameters to define a
    basic helix. Using just those parameters you can create simple single
    line helixes. But the primary purpose for helix() is to create a set
    of helical "wires" using non-zero values for taper_rpos, horz_offset and
    vert_offset to define solid helixes that can taper at each end to a point.
    This is useful for creating internal and external threads for nuts and
    bolts.  This is accomplished by invoking helix() multiple times with
    same radius, pitch, taper_rpos, inset_offset, first_t, and last_t.
    But with differing values for  horz_offset and vert_offset. And then
    using the returned functions to define the helical edges of the thread.

    :param radius: of the basic helix.
    :param pitch: of pitch of the helix per revolution. I.e the distance
        between the height of a single "turn" of the helix.
    :param height: of the cyclinder containing the helix.
    :param taper_out_rpos: is a decimal number with an inclusive range
        of 0..1 such that (taper_out_rpos * t_range) defines the
        t value where tapering out ends, it begins at t == first_t.
        A ValueError exception is raised if taper_out_rpos < 0 or > 1 or
        taper_out_rpos > taper_in_rpos. Default is 0 which is no out taper
    :param taper_in_rpos: is a decimal number with an inclusive range
        of 0..1 such that (taper_in_rpos * t_range) defines the
        t value where tapering in begins, it ends at t == last_t.
        A ValueError exception is raised if taper_out_rpos < 0 or > 1 or
        taper_out_rpos > taper_in_rpos. Default is 1 which is no in taper.
    :param inset_offset: the helix will start at z = inset_offset and will end
        at z = height - (2 * inset_offset). Default 0.
    :param first_t: is the first t value passed to the returned function. Default 0
    :param last_t: is the last t value passed to the returned function. Default 1
    """

    radius: float
    pitch: float
    height: float
    taper_out_rpos: float = 0
    taper_in_rpos: float = 1
    inset_offset: float = 0
    first_t: float = 0
    last_t: float = 1


@dataclass
class HelixLocation:
    # The internal thread radius and an array of HelixLocation
    radius: Optional[float] = None
    horz_offset: float = 0
    vert_offset: float = 0


def helix(
    h: Helix, hl: Optional[HelixLocation] = None
) -> Callable[[float], Tuple[float, float, float]]:
    """
    This function takes a Helix and optionally a HelixLocation which refines the
    location of the final helix. If HelixLocation is None then the radius is h.radius
    and and horz_offset and vert_offset will be 0. Since it's common for simple
    helixes HelixLocation.radius can also be None and h.radius will be used.

    This function returns a function, f. The funciton f that takes one parameter,
    an inclusive value between h.first_t and h.last_t.  We then define
    t_range=h.last_t-h.first_t and the rel_height=(h.last_t-t)/t_range. furthermore
    when using the h.first_t=0 and h.last_t=1, the defaults, t is rel_height.
    The value returned from f is a tuple(x, y, z) which defines a point
    on the helix defined by Helix.

    Credit: Adam Urbanczyk from cadquery [forum post](https://groups.google.com/g/cadquery/c/5kVRpECcxAU/m/7no7_ja6AAAJ)
    """
    if h.taper_out_rpos > h.taper_in_rpos:
        raise ValueError(
            f"taper_out_rpos:{h.taper_out_rpos} > taper_in_rpos:{h.taper_in_rpos}"
        )

    if h.taper_out_rpos < 0 or h.taper_out_rpos > 1:
        raise ValueError(f"taper_out_rpos:{h.taper_out_rpos} should be >= 0 and <= 1")

    if h.taper_in_rpos < 0 or h.taper_in_rpos > 1:
        raise ValueError(f"taper_in_rpos:{h.taper_in_rpos} should be >= 0 and <= 1")

    # Being "Tricky" to be flexible
    if hl is None:
        hl = HelixLocation(h.radius)
    elif hl.radius is None:
        hl.radius = h.radius

    # Reduce the height by 2 * inset_offset. Threads start at inset_offset
    # and end at height - inset_offset
    helix_height: float = (h.height - (2 * h.inset_offset))

    # The number or revolutions of the helix within the helix_height
    # set to 1 if pitch or helix_height is 0
    turns: float = h.pitch / helix_height if h.pitch != 0 and helix_height != 0 else 1

    # With this DISABLED points t_range will be negative
    # when h.last_t < h.first_t. This causes rel_height in "f"
    # to be negative and as a consequence # the values
    # generated by "f" will always be the same order.
    # See test_helix_backwards.
    #
    # If we ENABLE the code below and swap the order
    # if h.last_t < h.first_t then test_helix_backwards will
    # fail because the order of points in the resulting
    # array will be in reversed order.
    #
    # if h.last_t < h.first_t:
    #     h.last_t, h.first_t = h.first_t, h.last_t

    t_range: float = h.last_t - h.first_t

    taper_out_range: float = t_range * h.taper_out_rpos
    taper_out_ends: float = h.first_t + taper_out_range if taper_out_range > 0 else min(
        h.first_t, h.last_t
    )

    taper_in_range: float = t_range * (1 - h.taper_in_rpos)
    taper_in_starts: float = h.last_t - taper_in_range if taper_in_range > 0 else max(
        h.first_t, h.last_t
    )

    # print(f"helix: ft={h.first_t:.4f} lt={h.last_t:.4f} tr={t_range:.4f}")
    # print(f"helix: tor={taper_out_range:.4f} toe={taper_out_ends:.4f}")
    # print(f"helix: tir={taper_in_range:.4f} tis={taper_in_starts:.4f}")

    def func(t: float) -> Tuple[float, float, float]:
        """
        Return a tuple(x, y, z)
        :param t: A value between h.first_t .. h.last_t inclusive
        """

        taper_angle: float
        toffset: float = t - h.first_t
        rel_height: float = toffset / t_range if t_range != 0 else 0

        # print(f"f:  t={t:.4f}")
        # print(f"f:  tor={taper_out_range:.4f} toe={taper_out_ends:.4f}")
        # print(f"f:  tir={taper_in_range:.4f} tis={taper_in_starts:.4f}")
        if t < taper_out_ends:
            # Taper out from a point, taper_scale will be between 0 and 1
            # This code path is used when t < taper_out and this helix
            # will smoothly taper from a point as taper angle starts at 0
            # and increases to p/2.
            # print(f"f:  out t={t:.4f} < taper_out_ends:{taper_out_ends}")
            taper_angle = pi / 2 * (t - h.first_t) / taper_out_range
        elif t <= taper_in_starts:
            # No tapering, taper_scale == 1
            # print(f"f:  no  t={t:.4f} >= taper_out_ends:{taper_out_ends} <= taper_in_starts:{taper_in_starts}")
            taper_angle = pi / 2
        else:
            # This code path is used when t > taper_in_starts and the this helix
            # will smoothly taper to a point as taper angle starts at p/2
            # and decrease to 0.
            # print(f"f:  in  t={t:.4f} > taper_in_starts:{taper_in_starts}")
            taper_angle = pi / 2 * (h.last_t - t) / taper_in_range

        # print(f"taper_angle={taper_angle}")
        taper_scale: float = sin(taper_angle)

        r: float = hl.radius + (hl.horz_offset * taper_scale)
        a: float = (2 * pi / turns) * rel_height

        x: float = r * sin(-a)
        y: float = r * cos(a)
        z: float = (
            (helix_height * (rel_height if h.pitch != 0 else 1))
            + (hl.vert_offset * taper_scale)
            + h.inset_offset
        )

        result: Tuple[float, float, float] = (x, y, z)
        # print(f"f:  tpa={degrees(taper_angle):.4f} tps={taper_scale:.4f} r={r:.4f} a={degrees(a):.4f}")
        # print(f"f:  t={t:.4f} toffset={toffset:.4f} rh={rel_height:.4f} result={result}")
        return result

    return func



import ast
import sys
from math import isclose, sqrt
from typing import Callable, List, Tuple

import plotly.express as px
import pytest
from numpy import arange

from taperable_helix import Helix, HelixLocation, helix

# Default abs_tol
absolute_tol: float = 1e-6
relative_tol: float = 1e-9

# Data directory string
data_dir_str = "tests/data/"

X: int = 0
Y: int = 1
Z: int = 2


def xDist_3d(
    linePt1: Tuple[float, float, float], linePt2: Tuple[float, float, float]
) -> float:
    xDist: float = linePt1[X] - linePt2[X]
    # print(f"xDist_3d:+- xDist={xDist}")
    return xDist


def yDist_3d(
    linePt1: Tuple[float, float, float], linePt2: Tuple[float, float, float]
) -> float:
    yDist: float = linePt1[Y] - linePt2[Y]
    # print(f"yDist_3d:+- yDist={yDist}")
    return yDist


def zDist_3d(
    linePt1: Tuple[float, float, float], linePt2: Tuple[float, float, float]
) -> float:
    zDist: float = linePt1[Z] - linePt2[Z]
    # print(f"zDist_3d:+- zDist={zDist}")
    return zDist


def dist_3d(
    linePt1: Tuple[float, float, float], linePt2: Tuple[float, float, float]
) -> float:
    dist: float = sqrt(
        pow(xDist_3d(linePt1, linePt2), 2)
        + pow(yDist_3d(linePt1, linePt2), 2)
        + pow(zDist_3d(linePt1, linePt2), 2)
    )
    # print(f"dist_2d:+- dist={dist}")
    return dist


def isclose_tuple(
    v1: Tuple[float, ...],
    v2: Tuple[float, ...],
    rel_tol: float = relative_tol,
    abs_tol: float = absolute_tol,
) -> bool:
    # print(f"isclose_tuple: v1={v1} v2={v2}")
    return all(
        [isclose(a, b, rel_tol=rel_tol, abs_tol=abs_tol) for a, b in zip(v1, v2)]
    )


def isclose_points(
    result: List[Tuple[float, float, float]],
    expected: List[Tuple[float, float, float]],
    rel_tol: float = relative_tol,
    abs_tol: float = absolute_tol,
) -> bool:
    # print(f"isclose_points: result={result} expected={expected}")
    return all([isclose_tuple(v1, v2) for v1, v2 in zip(result, expected)])


def generate_points(
    f: Callable[[float], Tuple[float, float, float]],
    first_t: float,
    last_t: float,
    inc: float,
) -> List[Tuple[float, float, float]]:
    pts: List[Tuple[float, float, float]] = list(map(f, arange(first_t, last_t, inc)))
    pts.append(f(last_t))
    # print(f"generate_points: len(pts)={len(pts)} first_t={first_t} last_t={last_t} pts={pts}")
    return pts


def write_points(fname: str, points: List[Tuple[float, float, float]]) -> None:
    with open(fname + ".txt", "w") as f:
        for x, y, z in points:
            f.writelines(f"{x}, {y}, {z},\n")


def read_points(fname: str) -> List[Tuple[float, float, float]]:
    points: List[Tuple[float, float, float]]
    with open(fname + ".txt", "r") as f:
        points = [ast.literal_eval(line) for line in f]
    return points


def print_points(prompt: str, points: List[Tuple[float, float, float]]) -> None:
    print(f"{prompt}: ", end="")
    for x, y, z in points:
        print(f"({x:.4f}, {y:.4f}, {z:.4f})", end=", ")
    print("")


def view(
    title: str,
    points: List[Tuple[float, float, float]],
    first_t: float,
    last_t: float,
    inc: float,
) -> None:
    # print(f"view: points={points}")
    fig = px.line_3d(
        title=title,
        x=[x for x, _, _ in points],
        y=[y for _, y, _ in points],
        z=[z for _, _, z in points],
    )
    fig.layout.scene.camera.projection.type = "orthographic"
    fig.show()


def doit(
    func_name: str,
    points: List[Tuple[float, float, float]],
    first_t: float,
    last_t: float,
    inc: float,
    viewable: bool = False,
    generate: bool = False,
) -> None:
    fname: str = data_dir_str + func_name
    if generate:
        write_points(fname, points)
    if viewable:
        view(func_name, points, first_t, last_t, inc)
    expected = read_points(fname)
    assert isclose_points(points, expected)


def test_helix(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    radius = 1
    pitch = 1
    height = 1
    h = Helix(radius=radius, pitch=pitch, height=height)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A "helix" with all points equidistant from z-axis by the raidus
    assert all(
        [
            isclose(
                dist_3d((0, 0, pt[Z]), pt),
                radius,
                rel_tol=relative_tol,
                abs_tol=absolute_tol,
            )
            for pt in points
        ]
    )


def test_helix_backwards(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    radius = 1
    pitch = 1
    height = 1
    h = Helix(radius=radius, pitch=pitch, height=height)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # Validate the swapping of first_t and last_t creates the same helix
    h.first_t, h.last_t = h.last_t, h.first_t
    f = helix(h)
    # Generate the points, since we're starting at last_t we must use -inc
    # otherwise we'll only generate the last point
    points_backwards = generate_points(f, h.first_t, h.last_t, -inc)
    assert isclose_points(points, points_backwards)


def test_helix_torp_0pt1_tirp_0pt9_ho_0pt2(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.05
    radius = 1
    pitch = 1
    height = 1
    torp = 0.1
    tirp = 0.9
    ho = 0.2
    h = Helix(
        radius=radius,
        pitch=pitch,
        height=height,
        taper_out_rpos=torp,
        taper_in_rpos=tirp,
    )
    f = helix(h, HelixLocation(horz_offset=ho))
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A "helix" with a convergence factor of 0.1. The distance from the center line
    # to the first and last points should be the radius.
    assert isclose_tuple((0, radius, 0), points[0])
    assert isclose_tuple((0, radius, height), points[-1])

    # The distance to the second point and the penultimate point are > radius
    assert dist_3d(points[1], (0, 0, points[1][Z])) > radius
    assert dist_3d(points[-2], (0, 0, points[-2][Z])) > radius

    # All other points should be equal to radius + ho
    assert all(
        [
            isclose(
                dist_3d((0, 0, pt[Z]), pt),
                radius + ho,
                rel_tol=relative_tol,
                abs_tol=absolute_tol,
            )
            for pt in points[2:-2]
        ]
    )


def test_helix_trp_validity(view, generate):
    radius = 1
    pitch = 1
    height = 1

    h = Helix(radius=radius, pitch=pitch, height=height,)

    h.taper_out_rpos = -0.1
    h.taper_in_rpos = 0.9
    with pytest.raises(ValueError):
        helix(h)

    h.taper_out_rpos = 1.00000000000001
    h.taper_in_rpos = 0.9
    with pytest.raises(ValueError):
        helix(h)

    h.taper_out_rpos = 0.1
    h.taper_in_rpos = -0.00000001
    with pytest.raises(ValueError):
        helix(h)

    h.taper_out_rpos = 0.1
    h.taper_in_rpos = 1.00000001
    with pytest.raises(ValueError):
        helix(h)

    h.taper_out_rpos = 0
    h.taper_in_rpos = 0.9
    helix(h)

    h.taper_out_rpos = 0.9
    h.taper_in_rpos = 1
    helix(h)

    h.taper_out_rpos = 0.1
    h.taper_in_rpos = 0.7
    helix(h)

    h.taper_out_rpos = 0.7
    h.taper_in_rpos = 0.8
    helix(h)

    h.taper_out_rpos = 0.1
    h.taper_in_rpos = 0.9
    helix(h)


def test_radius_0(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    h = Helix(radius=0, pitch=1, height=1)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )
    # A vertical line starting at 0 ending at 1
    first_pt = (0, 0, 0)
    last_pt = (0, 0, 1)
    assert points[0] == first_pt
    assert points[-1] == last_pt
    assert all(
        [
            (pt[0], pt[1], 0) == (0, 0, 0) and (pt >= first_pt) and (pt <= last_pt)
            for pt in points
        ]
    )


def test_radius_0_ft_0_lt_0(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    h = Helix(radius=0, pitch=1, height=1, first_t=0, last_t=0)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A point at origin
    assert len(points) == 1
    assert points[0] == (0, 0, 0)


def test_radius_0_ft_neg_1_lt_pos_1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    h = Helix(radius=0, pitch=1, height=1, first_t=-1, last_t=1)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )
    # A vertical line starting at 0 ending at 1
    first_pt = (0, 0, 0)
    last_pt = (0, 0, 1)
    assert points[0] == first_pt
    assert points[-1] == last_pt
    assert all(
        [
            (pt[0], pt[1], 0) == (0, 0, 0) and (pt >= first_pt) and (pt <= last_pt)
            for pt in points
        ]
    )


def test_radius_0_ft_0_lt_neg_1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = -0.1
    h = Helix(radius=0, pitch=1, height=1, first_t=0, last_t=-1)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A vertical line starting at 1 ending at 0
    first_pt = (0, 0, 0)
    last_pt = (0, 0, 1)
    assert points[0] == first_pt
    assert points[-1] == last_pt
    assert all(
        [
            (pt[0], pt[1], 0) == (0, 0, 0) and (pt >= first_pt) and (pt <= last_pt)
            for pt in points
        ]
    )


def test_radius_0_ft_neg_2_lt_neg_1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    h = Helix(radius=0, pitch=1, height=1, first_t=-2, last_t=-1)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A vertical line starting at 1 ending at 0
    first_pt = (0, 0, 0)
    last_pt = (0, 0, 1)
    assert points[0] == first_pt
    assert points[-1] == last_pt
    assert all(
        [
            (pt[0], pt[1], 0) == (0, 0, 0) and (pt >= first_pt) and (pt <= last_pt)
            for pt in points
        ]
    )


def test_radius_0_ft_neg_1_lt_neg_2(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = -0.1
    h = Helix(radius=0, pitch=1, height=1, first_t=-1, last_t=-2)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A vertical line starting at 0 ending at 1
    first_pt = (0, 0, 0)
    last_pt = (0, 0, 1)
    assert points[0] == first_pt
    assert points[-1] == last_pt
    assert all(
        [
            (pt[0], pt[1], 0) == (0, 0, 0) and (pt >= first_pt) and (pt <= last_pt)
            for pt in points
        ]
    )


def test_radius_0_height_0(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    h = Helix(radius=0, pitch=1, height=0)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # All points at origin (0, 0, 0)
    assert all([(pt == (0, 0, 0)) for pt in points])


def test_radius_0_height_neg_1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    h = Helix(radius=0, pitch=1, height=-1)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A vertical line starting at 0 ending at -1
    first_pt = (0, 0, 0)
    last_pt = (0, 0, -1)
    assert points[0] == first_pt
    assert points[-1] == last_pt
    assert all(
        [
            (pt[0], pt[1], 0) == (0, 0, 0) and (pt <= first_pt) and (pt >= last_pt)
            for pt in points
        ]
    )


def test_radius_0_ft_pos_0_lt_neg_1_height_neg_1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = -0.1
    h = Helix(radius=0, pitch=1, height=-1, first_t=0, last_t=-1)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A vertical line starting at 0 ending at -1
    first_pt = (0, 0, 0)
    last_pt = (0, 0, -1)
    assert points[0] == first_pt
    assert points[-1] == last_pt
    assert all(
        [
            (pt[0], pt[1], 0) == (0, 0, 0) and (pt <= first_pt) and (pt >= last_pt)
            for pt in points
        ]
    )


def test_radius_0_io_0pt1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    inset = 0.1
    h = Helix(radius=0, pitch=1, height=1, inset_offset=inset,)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A vertical line starting at 0.1 ending at 0.9
    first_pt = (0, 0, inset)
    last_pt = (0, 0, 1 - inset)
    assert points[0] == first_pt
    assert points[-1] == last_pt
    assert all(
        [
            (pt[0], pt[1], 0) == (0, 0, 0) and (pt >= first_pt) and (pt <= last_pt)
            for pt in points
        ]
    )


def test_radius_0_ft_0_lt_0_io_neg_0pt1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    first_t = 0
    last_t = 0
    inc = 0.1
    inset = -0.1
    h = Helix(
        radius=0, pitch=1, height=1, inset_offset=inset, first_t=first_t, last_t=last_t
    )
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, first_t, last_t, inc, viewable=view, generate=generate,
    )

    # A point at origin (0, 0, inset)
    assert len(points) == 1
    assert points[0] == (0, 0, inset)


def test_pitch_0_height_0(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    radius = 1
    height = 0
    h = Helix(radius=radius, pitch=0, height=height)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A "circle" centered around the origin
    origin = (0, 0, 0)
    center = (0, 0, height)
    assert center == origin
    assert all(
        [
            (pt[Z] == height)
            and isclose(
                dist_3d(center, pt), radius, rel_tol=relative_tol, abs_tol=absolute_tol
            )
            for pt in points
        ]
    )


def test_pitch_0_height_1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    radius = 1
    height = 1
    h = Helix(radius=radius, pitch=0, height=height)
    f = helix(h)
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A "circle" centered around the height
    center = (0, 0, height)
    assert all(
        [
            (pt[Z] == height)
            and isclose(
                dist_3d(center, pt), radius, rel_tol=relative_tol, abs_tol=absolute_tol
            )
            for pt in points
        ]
    )


def test_pitch_0_ho_1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    radius = 1
    ho = 1
    h = Helix(radius=radius, pitch=0, height=0)
    f = helix(h, HelixLocation(horz_offset=ho))
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A "circle" centered at 0 with a radius = radius + ho
    center = (0, 0, 0)
    assert all(
        [
            (pt[Z] == 0)
            and isclose(
                dist_3d(center, pt),
                radius + ho,
                rel_tol=relative_tol,
                abs_tol=absolute_tol,
            )
            for pt in points
        ]
    )


def test_pitch_0_vo_1(view, generate):
    func_name: str = sys._getframe().f_code.co_name
    inc = 0.1
    radius = 1
    vo = 1
    h = Helix(radius=radius, pitch=0, height=0)
    f = helix(h, HelixLocation(vert_offset=vo))
    points = generate_points(f, h.first_t, h.last_t, inc)
    doit(
        func_name, points, h.first_t, h.last_t, inc, viewable=view, generate=generate,
    )

    # A "circle" centered around the horz_offset
    center = (0, 0, vo)
    assert all(
        [
            (pt[Z] == vo)
            and isclose(
                dist_3d(center, pt), radius, rel_tol=relative_tol, abs_tol=absolute_tol
            )
            for pt in points
        ]
    )



#!/usr/bin/env python

"""The setup script."""

from typing import List

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements: List[str] = []

setup_requirements: List[str] = [
    "pytest-runner",
]

test_requirements: List[str] = [
    "pytest>=3.7",
]

setup(
    author="Wink Saville",
    author_email="wink@saville.com",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Generate helixes that can optionally taper to a point at each end.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="taperable_helix",
    name="taperable_helix",
    packages=find_packages(include=["taperable_helix"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/winksaville/py-taperable-helix",
    version="0.7.0",
    zip_safe=False,
)
